#!/usr/bin/python3

import datetime
import os
import sys
import traceback
import argparse

libs_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                        '..', 'lib'))
if libs_dir not in sys.path:
    sys.path.append(libs_dir)

try:
    # api will be under $CHARM_SOURCE_DIR/lib during unit tests
    from charms.layer.graylog.api import GraylogApi
except ImportError:
    # api will be under $NAGIOS_PLUGIN_DIR at runtime
    from api import GraylogApi

STATE_OK = 0
STATE_WARNING = 1
STATE_CRITICAL = 2
STATE_UNKNOWN = 3

OVERALL_STATE = STATE_OK


def try_check(name, gapi_con, args, output):
    global OVERALL_STATE
    try:
        func = getattr(sys.modules[__name__], name)
        (ret, msg) = func(gapi_con, args)
    except Exception:
        ret = STATE_UNKNOWN
        msg = '{} failed: {}'.format(name, traceback.format_exc())
    if ret == STATE_CRITICAL:
        output['critical'].append(msg)
        OVERALL_STATE = STATE_CRITICAL
    elif ret == STATE_WARNING:
        output['warning'].append(msg)
        if OVERALL_STATE != STATE_CRITICAL:
            OVERALL_STATE = STATE_WARNING
    elif ret == STATE_OK:
        output['ok'].append(msg)
    else:
        output['unknown'].append(msg)
        if OVERALL_STATE == STATE_OK:
            OVERALL_STATE = STATE_UNKNOWN
    return OVERALL_STATE


def indexer_cluster_health(gapi_con, args):
    status = gapi_con.indexer_cluster_health()['status']
    msg = 'Indexer cluster health: {}'.format(status)
    if status == 'red':
        return STATE_CRITICAL, msg
    elif status == 'yellow':
        return STATE_WARNING, msg
    elif status == 'green':
        return STATE_OK, msg
    else:
        return STATE_UNKNOWN, msg


def journal_state(gapi_con, args):
    uncommitted_msgs = gapi_con.journal_state()['uncommitted_journal_entries']
    msg = 'Journal uncommitted messages: {}'.format(uncommitted_msgs)
    if uncommitted_msgs <= args.uncommitted_warning:
        return STATE_OK, msg
    elif uncommitted_msgs < args.uncommitted_critical:
        return STATE_WARNING, msg
    else:
        return STATE_CRITICAL, msg


def outstanding_notifications(gapi_con, args):
    notifications = gapi_con.notifications_get()
    msg = 'Outstanding notifications: {}'.format(notifications['total'])
    if notifications['total'] == 0:
        return STATE_OK, msg
    for n in notifications['notifications']:
        if n['severity'] == 'urgent':
            return STATE_CRITICAL, msg
    return STATE_WARNING, msg


def indexer_failures(gapi_con, args):
    since = datetime.datetime.utcnow() - datetime.timedelta(days=1)
    failures = gapi_con.indexer_failures_count(since.isoformat())['count']
    msg = 'Indexer failures: {}'.format(failures)
    if failures == 0:
        return STATE_OK, msg
    elif failures < 10:
        return STATE_WARNING, msg
    else:
        return STATE_CRITICAL, msg


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--uncommitted-critical', type=int, default=100)
    parser.add_argument('--uncommitted-warning', type=int, default=0)
    parser.add_argument('url', default='http://localhost:9001/api')
    return parser.parse_args()


def main():
    args = parse_args()
    output = {
        'critical': [],
        'warning': [],
        'ok': [],
        'unknown': [],
    }
    token = None
    with open('/var/lib/nagios/graylog-api-token', 'rb') as f:
        token = f.read().decode("utf-8")
    gapi_con = GraylogApi(args.url, token, 'token')
    for check in ('indexer_cluster_health', 'journal_state', 'outstanding_notifications', 'indexer_failures'):
        try_check(check, gapi_con, args, output)
    if output['critical']:
        print('CRITICAL: {}'.format('; '.join(output['critical'])))
    if output['warning']:
        print('WARNING: {}'.format('; '.join(output['warning'])))
    if output['unknown']:
        for e in output['unknown']:
            print('UNKNOWN: {}'.format(e))
    if output['ok']:
        print('OK: {}'.format('; '.join(output['ok'])))

    sys.exit(OVERALL_STATE)


if __name__ == '__main__':
    main()
