#
#    "             "
#  mmm   m   m   mmm   m   m
#    #   #   #     #   #   #
#    #   #   #     #   #   #
#    #   "mm"#     #   "mm"#
#    #             #
#  ""            ""
# This file is managed by Juju. Local changes will be lost.
import json
import os
import requests

# We are in a charm environment
if os.environ.get('JUJU_UNIT_NAME'):
    from charmhelpers.core import hookenv
    from charmhelpers.core.hookenv import log
    charm = True
else:
    charm = False


class GraylogApi:

    def __init__(self, base_url, username, password, token_name='graylog-api'):
        self.base_url = base_url
        if not base_url.endswith('/'):
            self.base_url += '/'
        self.username = username
        self.password = password
        self.auth = (self.username, self.password)
        self.token_name = token_name
        self.token = None
        self.input_types = None

    def request(self, path, method='GET', data={}, params=None):
        if path.startswith('/'):
            path = path[1:]
        url = '{}{}?pretty=true'.format(self.base_url, path)
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Accept-Encoding': '',
            'X-Requested-By': 'Graylog Charm'
        }
        if data:
            data = json.dumps(data, indent=True)
        tries = 0
        while tries < 3:
            tries += 1
            try:
                resp = requests.api.request(method, url, auth=self.auth,
                                            data=data, params=params,
                                            headers=headers, timeout=3)
                if resp.ok:
                    if method == 'DELETE':
                        return True
                    return json.loads(resp.content.decode('utf-8'))
                else:
                    msg = 'API error code: {}'.format(resp.status_code)
                    if charm:
                        log(msg)
                        hookenv.status_set('blocked', msg)
                    else:
                        print(msg)
            except Exception as ex:
                msg = 'Error calling graylog api: {}'.format(ex)
                if charm:
                    log(msg)
                else:
                    print(msg)

    def token_get(self, token_name=None, halt=False):
        if self.token:
            return self.token
        if not token_name:
            token_name = self.token_name
        if self.password == 'token':
            self.token = self.username
            self.auth = (self.token, 'token')
            return self.token
        url = 'users/{}/tokens'.format(self.username)
        resp = self.request(url)
        if not resp:
            return
        for token in resp['tokens']:
            if token['name'] == token_name:
                self.token = token['token']
                self.auth = (token['token'], 'token')
                return self.token

        # None found, let's create a new
        url = '{}/{}'.format(url, token_name)
        resp = self.request(url, method='POST')
        if not halt:
            return self.token_get(token_name=token_name, halt=True)
        else:
            if charm:
                hookenv.status_set('blocked', 'Cannot get API token.')
            return None  # Don't loop until we run out of memory.

    def index_set_get(self, index_id=None):
        if not self.token:
            self.token_get()
        url = 'system/indices/index_sets'
        if index_id:
            return self.request('{}/{}'.format(url, index_id))
        resp = self.request(url)
        if resp:
            return resp.get('index_sets', None)

    def index_set_update(self, index_id, data):
        if not self.token:
            self.token_get()
        url = 'system/indices/index_sets/{}'.format(index_id)
        return self.request(url, method='PUT', data=data)

    def log_input_get(self, input_id=None):
        if not self.token:
            self.token_get()
        url = 'system/inputs'
        if input_id:
            return self.request('{}/{}'.format(url, input_id))
        resp = self.request(url)
        if resp:
            return resp.get('inputs', None)

    def log_input_update(self, input_id=None, data=None):
        if not self.token:
            self.token_get()
        url = 'system/inputs/'
        if input_id:
            method = 'PUT'
            url += input_id
        else:
            method = 'POST'
        return self.request(url, method=method, data=data)

    def log_input_remove(self, input_id):
        if not self.token:
            self.token_get()
        url = 'system/inputs/{}'.format(input_id)
        return self.request(url, method='DELETE')

    def map_log_input_type(self, type):
        """
        Map specified int types to Gray's input type objects. Examples:

          "Beats" -> org.graylog.plugins.beats.BeatsInput
          "Raw TCP" -> org.graylog2.inputs.raw.tcp.RawTCPInput
          "Syslog UDP" -> org.graylog2.inputs.syslog.udp.SyslogUDPInput
        """
        if not self.input_types:
            url = 'system/inputs/types'
            result = self.request(url)
            if result is not None:
                self.input_types = result['types']
            else:
                return None
        for k, v in self.input_types.items():
            if v.lower().find(type.lower()) != -1:
                return k
            # For raw input types, map something like "Raw TCP" and "Raw UDP"
            # instead of requiring the exact name, "Raw/Plaintext TCP".
            if k.lower().find('.'.join(type.split()).lower()) != -1:
                return k
        return None

    def cluster_get(self):
        if not self.token:
            self.token_get()
        url = 'cluster'
        return self.request(url)

    def user_get(self, username):
        if not self.token:
            self.token_get()
        url = 'users/{}'.format(username)
        return self.request(url)

    def user_create(self, username, password, read_only=True):
        if not self.token:
            self.token_get()
        user = self.user_get(username)
        if user:
            return user
        url = 'users/'
        d = {
            'username': username,
            'password': password,
            'full_name': username,
            'email': '{}@local'.format(username),
            'permissions': [],
        }
        self.request(url, method='POST', data=d)
        return self.user_get(username)

    def user_permissions_clear(self, username):
        if not self.token:
            self.token_get()
        user = self.user_get(username)
        if not user:
            return
        url = 'users/{}/permissions'.format(username)
        self.request(url, method='DELETE')

    def user_permissions_set(self, username, permission):
        if not self.token:
            self.token_get()
        user = self.user_get(username)
        if not user:
            return
        url = 'users/{}/permissions'.format(username)
        self.request(url, method='PUT', data={'permissions': permission})

    def indexer_cluster_health(self):
        if not self.token:
            self.token_get()
        return self.request('system/indexer/cluster/health')

    def indexer_failures_count(self, since):
        if not self.token:
            self.token_get()
        return self.request('system/indexer/failures/count', params={'since': since})

    def journal_state(self):
        if not self.token:
            self.token_get()
        return self.request('system/journal')

    def notifications_get(self):
        if not self.token:
            self.token_get()
        return self.request('system/notifications')
