"""Set up and manage a pipeline for log data
extraction
"""

import re
from .utils import is_v2


class GraylogTitleSourceItems:

    def __init__(self, graylogapi=None):
        self.graylogapi = graylogapi

    def get(self):
        return self.graylogapi.request(self.url)

    def get_for_title(self, title):
        plines = self.get()
        matches = [p for p in plines if p['title'] == title]
        if len(matches) < 1:
            return None
        return matches[0]

    def post(self, title, source):
        return self.graylogapi.request(
            self.url, method='POST',
            data={'title': title, 'source': source})

    def delete(self, title):
        id_ = self.get_for_title(title)
        url = '{}/{}'.format(self.url, id_)
        return self.graylogapi.request(url, method='DELETE')

    def put(self, id_, title, source):
        url = '{}/{}'.format(self.url, id_)
        return self.graylogapi.request(
            url, method='PUT',
            data={'title': title, 'source': source})

    def set(self, title, source):
        """Set item data (creating it if doesn't exist)
        """
        item = self.get_for_title(title)
        if item:
            self.put(item["id"], title, source)
        else:
            self.post(title, source)


class GraylogPipelines(GraylogTitleSourceItems):
    url = '{}/system/pipelines/pipeline'.format(
        '/plugins/org.graylog.plugins.pipelineprocessor' if is_v2() else '')


class GraylogRules(GraylogTitleSourceItems):
    url = '{}/system/pipelines/rule'.format(
        '/plugins/org.graylog.plugins.pipelineprocessor' if is_v2() else '')


class GraylogStreams:
    def __init__(self, graylogapi=None):
        self.graylogapi = graylogapi

    def get_streams(self):
        resp = self.graylogapi.request('/streams')
        return resp['streams']

    def get_stream_id_for(self, title):
        """Get first streams id with given title, or None if not found
        :param title: Stream title
        :return: Stream ID
        """
        streams = self.get_streams()
        st = [o for o in streams if o['title'] == title]
        if len(st) < 1:
            return None
        return st[0]['id']

    def connect_pipeline(self, pipeline, stream):
        """Connect pipeline to stream. This op is safe to repeat
        :param pipeline title
        :param stream title
        """
        url = '{}/system/pipelines/connections/to_stream'.format(
            '/plugins/org.graylog.plugins.pipelineprocessor' if is_v2() else '')

        streamid = self.get_stream_id_for(stream)
        pipes = GraylogPipelines(self.graylogapi)
        pipeline = pipes.get_for_title(pipeline)
        if pipeline:
            self.graylogapi.request(url, method="POST",
                                    data={"stream_id": streamid,
                                          "pipeline_ids": [pipeline['id']]})


class LogExtractPipelineException(Exception):
    """Error handling log extraction
    """


_logextract_pipe_template = '''
pipeline "{}"
stage 0 match either
{}
end'''


class LogExtractPipeline:
    """Manage a pipeline with rules for log extraction
    The pipeline comprises a single stage which holds
    a list of rules
    """

    title = 'juju-logextract-pipeline'
    all_messages = "All messages"

    rule_title_pat = re.compile(r'rule "(.*)"')

    def __init__(self, graylogapi=None):
        self.graylogapi = graylogapi
        self.pipelines = GraylogPipelines(graylogapi)
        self.rules = GraylogRules(graylogapi)
        self.streams = GraylogStreams(graylogapi)

    def get_rules(self):
        pipe = self.pipelines.get_for_title(self.title)
        return pipe['stages'][0]['rules']

    def parse_rule_title(self, rule):
        lines = rule.strip().splitlines()
        match = self.rule_title_pat.search(lines[0])
        if not match:
            raise LogExtractPipelineException(
                "Couldn't find title in {}".format(rule)
            )
        return match.group(1)

    def set_rules(self, rules):
        titles = []
        for rule in rules:
            title = self.parse_rule_title(rule)
            self.rules.set(title, rule)
            titles.append(title)
        titlerefs = ['rule "{}";\n'.format(t) for t in titles]
        pipesrc = _logextract_pipe_template.format(self.title, ''.join(titlerefs))
        self.pipelines.set(self.title, pipesrc)
        self.streams.connect_pipeline(self.title, self.all_messages)
