import hashlib
import os
import re
import shutil
import time
import yaml
from urllib.parse import urlparse

from charms.reactive import hook, when, when_any, when_not, when_not_all, is_state, remove_state, set_state
from charms.reactive.helpers import data_changed
from charmhelpers.core import host, hookenv, unitdata
from charmhelpers.contrib.charmsupport import nrpe
import charms.leadership

from charms.layer import snap
from charms.layer.graylog import (
    GraylogApi,
    LogExtractPipeline,
    SNAP_NAME,
    get_api_port,
    get_api_url,
    is_v2,
    validate_api_uri
)


CONF_FILE = '/var/snap/graylog/common/server.conf'
SNAP_CONF_FILE = '/snap/graylog/current/etc/graylog/server/server.conf'
# /snap/graylog is a read-only squashfs so we use the initial settings
# and write out an override to SERVER_DEFAULT_CONF_FILE
SHIPPED_SNAP_SERVER_DEFAULT_CONF_FILE = '/snap/graylog/current/etc/default/graylog-server'
SERVER_DEFAULT_CONF_FILE = '/var/snap/graylog/current/default-graylog-server'
ELASTICSEARCH_DISCOVERY_PORT = '9300'
SERVICE_NAME = 'snap.graylog.graylog'


@hook('upgrade-charm')
def upgrade_charm():
    remove_state('graylog.configured')


@when('config.changed')
@when_not('config.changed.channel')
def update_config():
    """Allow new runtime config options to be processed.

    Reset flags to allow new runtime config options to be written. Deployment
    config options (e.g. 'channel') are handled separately.
    """
    remove_state('graylog.configured')


rotation_strategies = {
    'time': {
        'class': 'org.graylog2.indexer.rotation.strategies.TimeBasedRotationStrategy',
        'special': 'rotation_period',
        'juju-conf': 'index_rotation_period',
    },
    'size': {
        'class': 'org.graylog2.indexer.rotation.strategies.SizeBasedRotationStrategy',
        'special': 'max_size',
        'juju-conf': 'index_rotation_size',
    },
    'msg_count': {
        'class': 'org.graylog2.indexer.rotation.strategies.MessageCountRotationStrategy',
        'special': 'max_docs_per_index',
        'juju-conf': 'index_rotation_msg_count',
    }
}


@when_not('snap.installed.graylog')
def install_graylog():
    """Install the graylog snap.

    The snap layer will first try to install a valid 'graylog.snap' resource.
    If a valid resource is not found, the snap will be installed from the snap
    store according to the configured 'channel' option.
    """
    channel = hookenv.config('channel')

    hookenv.status_set('maintenance', 'Installing graylog snap')
    snap.install(SNAP_NAME, channel=channel)


@when('snap.installed.graylog')
@when('config.changed.channel')
def refresh_graylog():
    """Refresh the graylog snap.

    A change to the 'channel' config option may warrant a snap refresh. If so,
    stop the service, backup the current config, and refresh.
    """
    channel = hookenv.config('channel')
    cur_snap = new_snap = snap.get_installed_version(SNAP_NAME)

    if channel:
        host.service_stop(SERVICE_NAME)
        # backup config file using the current version as a suffix
        if os.path.exists(CONF_FILE):
            shutil.copy2(CONF_FILE, '{path}.{ext}'.format(path=CONF_FILE, ext=cur_snap))

        hookenv.status_set('maintenance', 'Refreshing graylog snap {}'.format(channel))
        snap.refresh(SNAP_NAME, channel=channel)
        new_snap = snap.get_installed_version(SNAP_NAME)
    else:
        hookenv.log("Cannot refresh snap when the 'channel' config option is missing")

    # When the snap version changes, use the config file from the snap and
    # adjust states to ensure appropriate re-config happens.
    if cur_snap != new_snap:
        shutil.copy2(SNAP_CONF_FILE, CONF_FILE)
        data_changed('elasticsearch.relation', None)
        data_changed('mongodb.uri', None)
        remove_state('graylog.configured')

    report_status()


@when('snap.installed.graylog')  # noqa: C901
@when_not('graylog.configured')
def configure_graylog():
    db = unitdata.kv()
    if not os.path.exists(CONF_FILE):
        hookenv.status_set('maintenance', 'Waiting for {}'.format(CONF_FILE))
        return
    conf = hookenv.config()

    admin_password = db.get('admin_password')
    password_secret = db.get('password_secret')
    if not (admin_password and password_secret):
        hookenv.status_set('maintenance', 'Waiting for leader passwords')
        return
    pw_hash = hashlib.sha256(admin_password.encode('utf-8')).hexdigest()
    set_conf('root_password_sha2', pw_hash)
    set_conf('password_secret', password_secret)

    api_port = get_api_port()
    if is_v2():
        if conf['web_listen_uri']:
            url = urlparse(conf['web_listen_uri'])
            webport = db.get('web_listen_port')

            if webport == api_port:
                msg = "Config 'web_listen_uri' is using port {} which conflicts with default REST port {}".format(
                    webport, api_port
                )
                hookenv.status_set('blocked', msg)
                return

            set_conf('web_listen_uri', conf['web_listen_uri'])
            if url.port != webport:
                if webport:
                    hookenv.close_port(webport)
                db.set('web_listen_port', url.port)
                hookenv.open_port(url.port)

        if conf['rest_transport_uri']:
            set_conf('rest_transport_uri', validate_api_uri(conf['rest_transport_uri']))

        if conf['web_endpoint_uri']:
            set_conf('web_endpoint_uri', validate_api_uri(conf['web_endpoint_uri']))
    else:
        if conf['web_listen_uri']:
            url = urlparse(conf['web_listen_uri'])
            set_conf('http_bind_address', url.netloc)
        if conf['rest_transport_uri']:
            set_conf('http_publish_uri', validate_api_uri(conf['rest_transport_uri']))
        if conf['web_endpoint_uri']:
            set_conf('http_external_uri', validate_api_uri(conf['rest_transport_uri']))

    if conf.get('logextraction-rules'):
        rules_config = yaml.safe_load(conf.get('logextraction-rules'))
        if rules_config and isinstance(rules_config, list):
            _configure_logextraction(rules_config)

    set_jvm_heap_size(conf['jvm_heap_size'])

    hookenv.application_version_set(snap.get_installed_version(SNAP_NAME))
    hookenv.open_port(api_port)

    remove_state('graylog_api.configured')
    set_state('graylog.configured')
    report_status()


@when('leadership.is_leader')
@when_not_all('leadership.set.admin_password', 'leadership.set.password_secret')
def generate_leader_passwords():
    if not is_state('leadership.set.admin_password'):
        admin_password = host.pwgen(18)
        charms.leadership.leader_set(admin_password=admin_password)
    if not is_state('leadership.set.password_secret'):
        password_secret = host.pwgen(96)
        charms.leadership.leader_set(password_secret=password_secret)


@when_any('leadership.changed.admin_password', 'leadership.changed.password_secret')
def get_leader_passwords():
    db = unitdata.kv()
    admin_password = charms.leadership.leader_get('admin_password')
    db.set('admin_password', admin_password)
    password_secret = charms.leadership.leader_get('password_secret')
    db.set('password_secret', password_secret)
    remove_state('graylog.configured')


@when('leadership.is_leader')
def appoint_master():
    master = charms.leadership.leader_get('master')
    if master and (master == hookenv.local_unit() or master in set(hookenv.expected_peer_units())):
        # A master already exists
        return
    # Appoint myself as master
    charms.leadership.leader_set(master=hookenv.local_unit())


def is_master():
    return charms.leadership.leader_get('master') == hookenv.local_unit()


@when('leadership.changed.master')
def set_master():
    set_conf('is_master', is_master())


@when('graylog.configured')  # noqa: C901
def report_status():
    beats_connected = is_state('beats.connected')
    beats_available = is_state('beat.setup')
    es_connected = is_state('elasticsearch.connected')
    es_available = is_state('elasticsearch.available')
    mongodb_connected = is_state('mongodb.connected')
    mongodb_available = is_state('mongodb.available')

    # Check for missing required applications
    missing_apps = []
    if not es_connected:
        missing_apps.append('elasticsearch')
    if not mongodb_connected:
        missing_apps.append('mongodb')

    if missing_apps:
        # Report blocked if we have missing apps
        msg = "Missing relation to: {}".format(', '.join(missing_apps))
        hookenv.status_set('blocked', msg)
    else:
        # Report ready/waiting based on current relation states
        waiting_apps = []
        ready_apps = []

        # Beats
        if beats_connected and not beats_available:
            waiting_apps.append('filebeat')
        elif beats_available:
            ready_apps.append('filebeat')

        # Elasticsearch
        if es_connected and not es_available:
            waiting_apps.append('elasticsearch')
        elif es_available:
            ready_apps.append('elasticsearch')

        # MongoDB
        if mongodb_connected and not mongodb_available:
            waiting_apps.append('mongodb')
        elif mongodb_available:
            ready_apps.append('mongodb')

        # Waiting apps take precedent over ready apps
        if waiting_apps:
            msg = "Waiting for: {}".format(', '.join(waiting_apps))
            hookenv.status_set('waiting', msg)
        else:
            # If no waiting_apps and we made it this far, we have ready_apps
            msg = "Ready with: {}".format(', '.join(ready_apps))
            hookenv.status_set('active', msg)


@when('graylog.configured')
@when('mongodb.available')
@when('elasticsearch.available')
@when_not('graylog.needs_restart')
@when_not('graylog_api.configured')
def configure_graylog_api(*discard):
    """Adjust states to ensure API dependents will be correctly configured."""
    remove_state('beat.setup')
    remove_state('graylog_index_sets.configured')
    remove_state('graylog_inputs.configured')
    remove_state('graylog_nagios.configured')
    set_state('graylog_api.configured')


@when('graylog.configured')  # noqa: C901
@when('graylog_api.configured')
@when_not('graylog_index_sets.configured')
def configure_index_sets(*discard):
    conf = hookenv.config()
    db = unitdata.kv()
    g = GraylogApi(base_url=get_api_url(),
                   username='admin',
                   password=db.get('admin_password'),
                   token_name='graylog-charm')
    index_sets = g.index_set_get()
    if index_sets is None:
        return
    index_shards = conf['index_shards']
    if index_shards == 0:
        # Automatically work out the index shards based on how
        # many Elasticsearch units.
        for relid in hookenv.relation_ids('elasticsearch'):
            index_shards += len(hookenv.related_units(relid))
        if not index_shards:
            hookenv.log("Can't work out number of elasticsearch units")
            return

    for iset in index_sets:
        log = ''
        if iset['shards'] != index_shards:
            log += ' shards from {} to {}'.format(iset['shards'], index_shards)
        iset['shards'] = index_shards
        if iset['replicas'] != conf['index_replicas']:
            log += ' replicas from {} to {}'.format(iset['replicas'], conf['index_replicas'])
        iset['replicas'] = conf['index_replicas']
        if conf['index_rotation_strategy'] in rotation_strategies:
            rs = rotation_strategies[conf['index_rotation_strategy']]
            if iset['rotation_strategy_class'] != rs['class']:
                log += ' rotation strategy from {} to {} ({}={})'.format(
                    iset['rotation_strategy_class'], rs['class'], rs['special'], conf[rs['juju-conf']]
                )
            iset['rotation_strategy_class'] = rs['class']
            iset['rotation_strategy'] = {
                'type': '{}Config'.format(rs['class']),
                rs['special']: conf[rs['juju-conf']],
            }
        if iset['retention_strategy']['max_number_of_indices'] != conf['index_retention_count']:
            log += ' retention max_number_of_indices={}'.format(conf['index_retention_count'])
        iset['retention_strategy_class'] = 'org.graylog2.indexer.retention.strategies.DeletionRetentionStrategy'
        iset['retention_strategy'] = {
            'type': 'org.graylog2.indexer.retention.strategies.DeletionRetentionStrategyConfig',
            'max_number_of_indices': conf['index_retention_count'],
        }
        if not g.index_set_update(iset['id'], iset):
            return
        if log:
            hookenv.log('Updated{}'.format(log))

    set_state('graylog_index_sets.configured')


def _configure_logextraction(rules_config):
    db = unitdata.kv()
    g = GraylogApi(base_url=get_api_url(),
                   username='admin',
                   password=db.get('admin_password'),
                   token_name='graylog-charm')
    logext = LogExtractPipeline(g)
    logext.set_rules(rules_config)


def _check_input_exists(current_inputs, new):
    """
    Given the current list of configured Graylog inputs, check if new input
    exists and return an ID as well as "changed" flag to indicate that the
    new input provided has changed (input type, bind address, and port).
    """
    changed = True
    cur = None
    for input in current_inputs:
        if input['title'] == new['title']:
            cur = input
            break
    if not cur:
        return (True, None)
    current = [cur['type'], cur['attributes']['bind_address'], str(cur['attributes']['port'])]
    new = [new['type'], new['configuration']['bind_address'], str(new['configuration']['port'])]
    if sorted(new) == sorted(current):
        changed = False
    return (changed, cur['id'])


def _close_uneeded_ports(current_inputs, new):
    for cur in current_inputs:
        title = cur['title']
        # ignore Custom inputs from web interface
        if title.lower().startswith('custom'):
            continue
        # ignore juju_beats inputs from elastic-beats relation
        if title.lower() == 'juju_beats_input':
            continue
        type = cur['type']
        if 'udp' in type.lower():
            proto = 'UDP'
        else:
            proto = 'TCP'
        port = cur['attributes']['port']
        if '{}/{}'.format(port, proto) not in new:
            hookenv.close_port(port, proto)


def _remove_old_inputs(inputs, new_inputs):
    """
    Removes any old/previously defined inputs, or any unknown. We'll keep those
    prefixed with 'Custom' to allow configuring additional inputs via the web
    UI.
    """
    db = unitdata.kv()
    g = GraylogApi(base_url=get_api_url(),
                   username='admin',
                   password=db.get('admin_password'),
                   token_name='graylog-charm')
    success = True
    for cur in inputs:
        title = cur['title']
        # ignore Custom inputs from web interface
        if title.lower().startswith('custom'):
            continue
        # ignore juju_beats inputs from elastic-beats relation
        if title.lower().startswith('juju_beats_'):
            continue
        if title in new_inputs:
            continue
        if g.log_input_remove(cur['id']):
            hookenv.log('Removed old input: {} ({})'.format(cur['title'], cur['id']))
        else:
            success = False
    return success


@when('beat.setup')
@when_not('beats.connected')
def graylog_remove_filebeat_input():
    '''Remove the logstash beats client.'''
    hookenv.log("Removing filebeats input")

    db = unitdata.kv()
    g = GraylogApi(base_url=get_api_url(),
                   username='admin',
                   password=db.get('admin_password'),
                   token_name='graylog-charm')

    inputs = g.log_input_get()

    if inputs is not None:

        for inp in inputs:

            if inp['title'].lower() == "juju_beats_input":

                g.log_input_remove(inp['id'])
                hookenv.log('Removed beats input: {} ({})'.format(inp['title'], inp['id']))
                hookenv.status_set('active', 'Removed beats input {}'.format(inp['title']))
    remove_state('beat.setup')


@when('graylog.configured')
@when('graylog_api.configured')
@when('beats.connected')
@when_not('beat.setup')
def setup_beats_input(filebeat):
    port = hookenv.config('beats_port')
    hookenv.log("Setting up beats input on port {}".format(port))
    filebeat.provide_data(port)

    db = unitdata.kv()
    g = GraylogApi(base_url=get_api_url(),
                   username='admin',
                   password=db.get('admin_password'),
                   token_name='graylog-charm')

    type = g.map_log_input_type('Beats')
    if type is None:
        return

    d = {
        'title': "juju_beats_input",
        'type': type,
        'global': "true",
        'configuration': {
            'bind_address': '0.0.0.0',
            'port': port,
            'tcp_keepalive': 'true'
        }
    }

    proto = 'TCP'
    hookenv.open_port(port, proto)
    inputs = g.log_input_get()

    (changed, input_id) = _check_input_exists(inputs, d)
    if changed:
        if input_id:
            hookenv.log("updating beats input with new port {}".format(port))
        else:
            hookenv.log("configuring beats input on port {}".format(port))
        ret = g.log_input_update(input_id, d)
        if not ret:
            return
    else:
        hookenv.log("beats input already configured")

    set_state('beat.setup')


@when('graylog.configured')
@when('graylog_api.configured')
@when_not('graylog_inputs.configured')
def configure_inputs(*discard):
    """
    Configure log inputs in Graylog via the API.
    """
    conf = hookenv.config()
    db = unitdata.kv()
    g = GraylogApi(base_url=get_api_url(),
                   username='admin',
                   password=db.get('admin_password'),
                   token_name='graylog-charm')
    inputs = g.log_input_get()
    hookenv.log("Inputs: {}".format(inputs))
    if inputs is None:
        hookenv.log("Not configuring inputs.  GraylogApi returned None")
        return
    new_opened_ports = []
    new_inputs = []
    for new in yaml.safe_load(conf['log_inputs']) or {}:
        type = g.map_log_input_type(new['type'])
        if not type:
            hookenv.log('Input type "{}" not supported'.format(new['type']))
            continue
        d = {
            'title': new['name'],
            'type': type,
            'global': "true",
            'configuration': {
                'bind_address': new['bind_address'],
                'port': new['bind_port'],
            }
        }
        if 'udp' in type.lower():
            proto = 'UDP'
        else:
            proto = 'TCP'
            d['configuration']['tcp_keepalive'] = 'true'

        hookenv.open_port(new['bind_port'], proto)
        new_opened_ports.append('{}/{}'.format(new['bind_port'], proto))

        (changed, input_id) = _check_input_exists(inputs, d)
        if not changed:
            new_inputs.append(new['name'])
            continue
        if input_id:
            change_text = 'Updated existing input'
        else:
            change_text = 'Adding new input'
        ret = g.log_input_update(input_id, d)
        if not ret:
            # Can't add input, let's try again later
            return
        hookenv.log('{}: {} ({})'.format(change_text, new['name'], ret['id']))
        new_inputs.append(new['name'])

    # Now remove inputs that should no longer be there as well as close ports
    _close_uneeded_ports(inputs, new_opened_ports)
    refreshed_inputs = g.log_input_get()
    if not _remove_old_inputs(refreshed_inputs, new_inputs):
        # Can't delete input, let's try again later
        return
    set_state('graylog_inputs.configured')


@when('graylog.configured')
@when('elasticsearch.available')
def configure_elasticsearch_connection(elasticsearch):
    cluster = ""
    # Pre-2.3.x discovery unicast hosts (binary ElasticSearch protocol)
    discovery_hosts = []
    # 2.3.x and above HTTP REST API
    http_hosts = []
    for unit in elasticsearch.list_unit_data():
        cluster_name = unit['cluster_name']
        if cluster_name is not None:
            cluster = cluster_name
        http_hosts.append('http://{}:{}'.format(unit['host'], unit['port']))
        discovery_hosts.append('{}:{}'.format(unit['host'], ELASTICSEARCH_DISCOVERY_PORT))

    if not cluster:
        conf = hookenv.config()
        cluster = conf['elasticsearch_cluster_name']

    if not data_changed('elasticsearch.relation', {'cluster_name': cluster, 'hosts': http_hosts}):
        return

    set_conf('elasticsearch_cluster_name', cluster)
    set_conf('elasticsearch_discovery_zen_ping_unicast_hosts ', ', '.join(discovery_hosts))
    set_conf('elasticsearch_hosts', ', '.join(http_hosts))
    # Elastic search does not reliably pick the right ip to listen on
    set_conf('elasticsearch_network_host', hookenv.unit_private_ip())

    hookenv.log('Updated elasticsearch configuration')
    report_status()


@when('graylog.configured')
@when('mongodb.available')
@when_not('mongodb.removed')
def configure_mongodb_connection(mongodb):
    mongodb_hosts = []
    for mongo_host in mongodb.connection_strings():
        mongodb_hosts.append(mongo_host)

    mongodb_uri = 'mongodb://{}/graylog'.format(','.join(mongodb_hosts))
    if not data_changed('mongodb.uri', mongodb_uri):
        return

    set_conf('mongodb_uri', mongodb_uri)
    hookenv.log('Updated mongodb configuration')
    report_status()


@when('graylog.configured')
@when('website.available')
def update_reverseproxy_config(website):
    services = ""
    conf = hookenv.config()
    if conf['web_listen_uri']:
        url = urlparse(conf['web_listen_uri'])
        website.configure(port=url.port)
        services += "- {service_name: web, service_port: " + str(url.port) + "}\n"
    if is_v2():
        # NB: there is no revproxy to the api endpoint in v3+.
        services += "- {service_name: api, service_port: " + get_api_port() + "}\n"
    website.set_remote(all_services=services)


@when('graylog.configured')
@when('graylog.needs_restart')
def restart_service(service=SERVICE_NAME):
    """ Restart the service and on failure wait 15 seconds and try again.
        This handles situations when relations are adding quick enough to trigger
        systemd's standard protection against frequent restarts.
    """
    host.service_restart(service)
    remove_state('graylog.needs_restart')
    if host.service_running(service):
        return

    time.sleep(15)
    host.service_restart(service)


def set_conf(key, value, conf_path=CONF_FILE):
    """ Updates the graylog configuration setting lines matching 'key = value'.
        The configuration contains randomly generated values specific to each install
        so setting the configuration this way avoids having to regenerate those values.
    """
    conf = []
    key_re = re.compile('^#*\\s*{}'.format(key))

    changed = False
    replaced = False
    with open(conf_path, 'rb') as conf_file:
        found = False
        for line in conf_file:
            line = line.decode().rstrip()
            if not key_re.match(line):
                conf.append(line)
                continue
            found = True
            if not replaced:
                new = '{} = {}'.format(key, value).rstrip()
                if line != new:
                    hookenv.log('Updating configuration file option "{}".'.format(key))
                    changed = True
                conf.append(new)
                replaced = True
            elif not line.startswith('#'):
                conf.append('#' + line)
            else:
                conf.append(line)
        if not found:
            hookenv.log('Updating configuration file new option "{}".'.format(key))
            changed = True
            conf.append('{} = {}'.format(key, value))

    if changed:
        set_state('graylog.needs_restart')
        with open(conf_path, 'wb') as conf_file:
            conf_file.write('\n'.join(conf).encode())
        return True

    return False


def parse_java_opts(opts, size):
    foundXms = False
    foundXmx = False
    new_opts = []
    for opt in opts.split():
        if opt.startswith('-Xms'):
            foundXms = True
            if size == '':
                continue
            new_opts.append('-Xms{}'.format(size))
        elif opt.startswith('-Xmx'):
            foundXmx = True
            if size == '':
                continue
            new_opts.append('-Xmx{}'.format(size))
        else:
            new_opts.append(opt)
    if not foundXms:
        new_opts.append('-Xms{}'.format(size))
    if not foundXmx:
        new_opts.append('-Xmx{}'.format(size))
    return new_opts


def set_jvm_heap_size(heap_size='1G', conf_path=SERVER_DEFAULT_CONF_FILE):
    size = heap_size.lower()
    conf = []
    key_re = re.compile('^\\s*GRAYLOG_SERVER_JAVA_OPTS\\s*="(.*)"$')
    changed = False
    # If specified /etc/default/graylog-server doesn't exist, use the default
    # file shipped out by the snap
    if not os.path.exists(conf_path):
        updated_conf_path = SHIPPED_SNAP_SERVER_DEFAULT_CONF_FILE
    else:
        updated_conf_path = conf_path
    with open(updated_conf_path, 'rb') as conf_file:
        for line in conf_file:
            line = line.decode()
            m = key_re.match(line)
            if not m:
                conf.append(line)
                continue
            new_opts = parse_java_opts(m.group(1), size)
            new_line = 'GRAYLOG_SERVER_JAVA_OPTS="{}"\n'.format(' '.join(new_opts))
            if new_line != line:
                hookenv.log('Updating GRAYLOG_SERVER_JAVA_OPTS="{}".'.format(''.join(new_opts)))
                changed = True
            conf.append(new_line)

    if changed:
        set_state('graylog.needs_restart')
        with open(conf_path, 'wb') as conf_file:
            conf_file.write(''.join(conf).encode())
        return True
    return False


@when('graylog.configured')
@when('nrpe-external-master.available')
@when_not('graylog_nagios.configured')
def configure_nagios(nagios):
    db = unitdata.kv()
    nagios_password = db.get('nagios_password')
    if not nagios_password:
        nagios_password = host.pwgen(18)
        db.set('nagios_password', nagios_password)

    api_port = get_api_port()
    api_url = get_api_url()

    # Ask charmhelpers.contrib.charmsupport's nrpe to work out our hostname
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname, primary=True)

    conf = hookenv.config()
    if conf['web_listen_uri']:
        url = urlparse(conf['web_listen_uri'])
        # LP:#1706265: Work around this bug by escaping some characters.
        check_string = '\\<title\\>Graylog\\ Web\\ Interface\\</title\\>'
        url_path = url.path
        # Catch URI without an explicit path
        if not url_path:
            url_path = '/'
        # Ensure an ending slash
        if not url_path.endswith('/'):
            url_path += '/'
        nrpe_setup.add_check(
            'graylog_http',
            'Graylog Web UI check',
            '/usr/lib/nagios/plugins/check_http -I {} -p {} -u "{}" -s "{}"'
            .format(url.hostname, str(url.port), url_path, check_string)
        )
    check_string = 'cluster_id'
    nrpe_setup.add_check(
        'graylog_api',
        'Greylog API check',
        '/usr/lib/nagios/plugins/check_http -I 127.0.0.1 -p {} -u /api -s "{}"'.format(api_port, check_string)
    )

    # For nagios checks via API, we need to create a user and token
    g = GraylogApi(base_url=api_url,
                   username='admin',
                   password=db.get('admin_password'),
                   token_name='graylog-charm')
    g.user_create('nagios', nagios_password)
    g.user_permissions_set('nagios', ['users:tokenlist', 'users:tokencreate',
                                      'indexercluster:read', 'indices:failures',
                                      'notifications:read', 'journal:read'])
    g = GraylogApi(base_url=api_url,
                   username='nagios',
                   password=nagios_password,
                   token_name='nagios')
    nagios_token = g.token_get()
    if not nagios_token:
        # API not ready, let's write out what we have anyways.
        nrpe_setup.write()
        hookenv.log("No nagios token, return from setup early")
        return
    host.write_file('/var/lib/nagios/graylog-api-token', nagios_token.encode(),
                    group='nagios', perms=0o640)
    nagios_plugins = '/usr/local/lib/nagios/plugins/'
    for f in ('files/check_graylog_health.py', 'lib/charms/layer/graylog/api.py'):
        with open(f, 'rb') as content:
            fname = os.path.join(nagios_plugins, os.path.basename(f))
            host.write_file(fname, content.read(), perms=0o755)
    nrpe_setup.add_check(
        'graylog_health',
        'Graylog Health check',
        '{} --uncommitted-warning {} --uncommitted-critical {} {}'.format(
            os.path.join(nagios_plugins, 'check_graylog_health.py'),
            conf.get('nagios_uncommitted_warn', 0),
            conf.get('nagios_uncommitted_crit', 100),
            api_url)
    )

    nrpe_setup.write()
    set_state('graylog_nagios.configured')
