import os
import sys
import tempfile
import unittest
from unittest import mock
from charms.reactive import set_state, is_state

# some dep layers only exists in the built charm; mock them out before
# the graylog imports since those depend on post-build charms.* layers
leader_mock = mock.Mock()
sys.modules['charms.leadership'] = leader_mock
snap_mock = mock.Mock()
sys.modules['charms.layer.snap'] = snap_mock

from reactive.graylog import (
    GraylogApi,
    configure_graylog,
    install_graylog,
    refresh_graylog,
    set_conf,
    set_jvm_heap_size,
    update_config,
    upgrade_charm,
    _check_input_exists)  # noqa: E402
from files import check_graylog_health  # noqa: E402


class TestInstallUpdateUpgrade(unittest.TestCase):
    """Test that install / update / upgrade clears our configured flag."""

    @mock.patch('reactive.graylog.snap')
    @mock.patch('reactive.graylog.hookenv.config')
    def test_install(self, mock_config, mock_snap):
        class Config(dict):
            def __init__(self, *args, **kw):
                super(Config, self).__init__(*args, **kw)

        mock_config.return_value = Config({'channel': '2/stable'})
        install_graylog()
        self.assertTrue(mock_snap.install.called)
        self.assertFalse(is_state('graylog.configured'))

    def test_update(self):
        set_state('graylog.configured')
        update_config()
        self.assertFalse(is_state('graylog.configured'))

    def test_upgrade(self):
        set_state('graylog.configured')
        upgrade_charm()
        self.assertFalse(is_state('graylog.configured'))


class TestRefresh(unittest.TestCase):
    """Test the snap refresh handler."""

    @mock.patch('reactive.graylog.snap')
    @mock.patch('reactive.graylog.hookenv.config')
    def test_no_channel(self, mock_config, mock_snap):
        """Ensure a no-op if there is no 'channel' config."""
        mock_config.return_value = ''
        mock_snap.get_installed_version.return_value = '2.foo'
        set_state('graylog.configured')
        refresh_graylog()
        self.assertFalse(mock_snap.refresh.called)
        self.assertTrue(is_state('graylog.configured'))

    @mock.patch('reactive.graylog.shutil')
    @mock.patch('reactive.graylog.os')
    @mock.patch('reactive.graylog.host')
    @mock.patch('reactive.graylog.snap')
    @mock.patch('reactive.graylog.hookenv.config')
    def test_no_change(self, mock_config, mock_snap, mock_host, mock_os, mock_shutil):
        """Ensure no reconfiguration if the snap version does not change."""
        mock_config.return_value = 'foo'
        mock_snap.get_installed_version.side_effect = ['2.foo', '2.foo']
        mock_os.path.exists.return_value = True
        mock_shutil.copy2.return_value = 'foo'
        set_state('graylog.configured')
        refresh_graylog()
        self.assertTrue(mock_snap.refresh.called)
        self.assertTrue(is_state('graylog.configured'))

    @mock.patch('reactive.graylog.data_changed')
    @mock.patch('reactive.graylog.shutil.copy2')
    @mock.patch('reactive.graylog.os')
    @mock.patch('reactive.graylog.host')
    @mock.patch('reactive.graylog.snap')
    @mock.patch('reactive.graylog.hookenv.config')
    def test_change(self, mock_config, mock_snap, mock_host, mock_os, mock_shutil, mock_data_changed):
        """Ensure calls and states when the snap has changed."""
        mock_config.return_value = 'foo'
        mock_snap.get_installed_version.side_effect = ['2.foo', '3.foo']
        mock_os.path.exists.return_value = True
        mock_shutil.return_value = 'foo'
        set_state('graylog.configured')
        refresh_graylog()
        self.assertTrue(mock_host.service_stop.called)
        self.assertTrue(mock_snap.refresh.called)
        self.assertEqual(mock_shutil.call_count, 2)
        self.assertEqual(mock_data_changed.call_count, 2)
        self.assertFalse(is_state('graylog.configured'))


class TestConfig(unittest.TestCase):
    """Test the configure_graylog handler."""

    @mock.patch('reactive.graylog.hookenv')
    @mock.patch('reactive.graylog.os')
    def test_no_conf_file(self, mock_os, mock_hookenv):
        """Ensure a status message if there is no config file present."""
        mock_os.path.exists.return_value = False
        configure_graylog()
        self.assertTrue(mock_hookenv.status_set.called)

    @mock.patch('reactive.graylog.set_jvm_heap_size')
    @mock.patch('reactive.graylog.validate_api_uri')
    @mock.patch('reactive.graylog.is_v2')
    @mock.patch('reactive.graylog.set_conf')
    @mock.patch('reactive.graylog.hookenv')
    @mock.patch('reactive.graylog.os')
    @mock.patch('reactive.graylog.unitdata')
    def test_v2_config(self, mock_ud, mock_os, mock_hookenv, mock_set_conf, mock_v2, mock_validate_uri, mock_jvm_heap):
        """Ensure calls and states with various config values for v2."""
        class Config(dict):
            def __init__(self, *args, **kw):
                super(Config, self).__init__(*args, **kw)

        mock_ud.kv.return_value = {'admin_password': 'password', 'password_secret': 'secret'}
        mock_os.path.exists.return_value = True
        mock_hookenv.config.return_value = Config({'jvm_heap_size': '1G',
                                                   'web_listen_uri': None,
                                                   'rest_transport_uri': 'foo/api',
                                                   'web_endpoint_uri': 'bar/api'})
        mock_v2.return_value = True
        configure_graylog()
        self.assertEqual(mock_validate_uri.call_count, 2)
        self.assertTrue(mock_jvm_heap.called)
        self.assertTrue(mock_hookenv.application_version_set.called)
        self.assertTrue(mock_hookenv.open_port.called)

    @mock.patch('reactive.graylog.set_jvm_heap_size')
    @mock.patch('reactive.graylog.validate_api_uri')
    @mock.patch('reactive.graylog.is_v2')
    @mock.patch('reactive.graylog.set_conf')
    @mock.patch('reactive.graylog.hookenv')
    @mock.patch('reactive.graylog.os')
    @mock.patch('reactive.graylog.unitdata')
    def test_v3_config(self, mock_ud, mock_os, mock_hookenv, mock_set_conf, mock_v2, mock_validate_uri, mock_jvm_heap):
        """Ensure calls and states with various config values for v3."""
        class Config(dict):
            def __init__(self, *args, **kw):
                super(Config, self).__init__(*args, **kw)

        mock_ud.kv.return_value = {'admin_password': 'password', 'password_secret': 'secret'}
        mock_os.path.exists.return_value = True
        mock_hookenv.config.return_value = Config({'jvm_heap_size': '1G',
                                                   'web_listen_uri': None,
                                                   'rest_transport_uri': 'foo/api',
                                                   'web_endpoint_uri': 'bar/api'})
        mock_v2.return_value = False
        configure_graylog()
        self.assertEqual(mock_validate_uri.call_count, 2)
        self.assertTrue(mock_jvm_heap.called)
        self.assertTrue(mock_hookenv.application_version_set.called)
        self.assertTrue(mock_hookenv.open_port.called)


initial_conf = u"""#key1 = value1
key2 = value2
key2=value3
### key3   =  value 4
key4 = value5"""

update1_conf = u"""#key1 = value1
key2 = value2
key2=value3
### key3   =  value 4
key4 = valueUpdated"""

commented1_conf = u"""key1 = value1
key2 = value2
key2=value3
### key3   =  value 4
key4 = value5"""

commented2_conf = u"""key1 = value1
key2 = value2
key2=value3
key3 = value4
key4 = value5"""

duplicate1_conf = u"""#key1 = value1
key2 = value6
#key2=value3
### key3   =  value 4
key4 = value5"""

log_input_types = {
    "org.graylog.plugins.beats.BeatsInput": "Beats",
    "org.graylog2.inputs.raw.udp.RawUDPInput": "Raw/Plaintext UDP",
}

log_inputs = [{
    "title": "Beats Input",
    "global": "true",
    "name": "Beats",
    "created_at": "2017-08-17T10:19:10.634Z",
    "type": "org.graylog.plugins.beats.BeatsInput",
    "creator_user_id": "admin",
    "attributes": {
        "bind_address": "0.0.0.0",
        "port": 11044
    },
    "id": "5994f3aab0d37f23c38adbda"
}]


class TestSetConf(unittest.TestCase):

    def setUp(self):
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        temp_file.write(initial_conf.encode())
        temp_file.close()
        self.conf_file = temp_file.name

    def tearDown(self):
        os.remove(self.conf_file)

    def conf_equals(self, want):
        with open(self.conf_file, 'r') as conf:
            got = conf.read()
            if got == want:
                return True
            print('{}\n != \n{}'.format(got, want))
        return False

    def test_update(self):
        set_conf('key4', 'value5', self.conf_file)
        self.assertTrue(self.conf_equals(initial_conf))

        set_conf('key4', 'valueUpdated', self.conf_file)
        self.assertTrue(self.conf_equals(update1_conf))

    def test_set_commented(self):
        set_conf('key1', 'value1', self.conf_file)
        self.assertTrue(self.conf_equals(commented1_conf))

        set_conf('key3', 'value4', self.conf_file)
        self.assertTrue(self.conf_equals(commented2_conf))

    def test_handle_duplicates(self):
        set_conf('key2', 'value6', self.conf_file)
        self.assertTrue(self.conf_equals(duplicate1_conf))

    def test_no_changed(self):
        self.assertFalse(set_conf('key2', 'value2', self.conf_file))

    def test_changed(self):
        self.assertTrue(set_conf('key2', 'some-new-value', self.conf_file))

    def test_input_type_exists(self):
        g = GraylogApi('', '', '')
        g.input_types = log_input_types
        self.assertTrue(g.map_log_input_type('beats'))

    def test_input_type_raw(self):
        g = GraylogApi('', '', '')
        g.input_types = log_input_types
        self.assertTrue(g.map_log_input_type('raw udp'))

    def test_input_type_invalid(self):
        g = GraylogApi('', '', '')
        g.input_types = log_input_types
        self.assertFalse(g.map_log_input_type('some-weird-type'))

    def test_input_exists(self):
        new_input = {
            "title": "Beats Input",
            "type": "org.graylog.plugins.beats.BeatsInput",
            "configuration": {
                "bind_address": "0.0.0.0",
                "port": 11044
            }
        }
        (changed, input_id) = _check_input_exists(log_inputs, new_input)
        self.assertTrue(input_id == "5994f3aab0d37f23c38adbda" and changed is False)

    def test_input_changed(self):
        new_input = {
            "title": "Beats Input",
            "type": "org.graylog.plugins.beats.BeatsInput",
            "configuration": {
                "bind_address": "0.0.0.0",
                "port": 12222
            }
        }
        (changed, input_id) = _check_input_exists(log_inputs, new_input)
        self.assertTrue(changed is True)

    def test_input_new(self):
        new_input = {
            "title": "Beats Input 2",
            "type": "org.graylog.plugins.beats.BeatsInput",
            "configuration": {
                "bind_address": "0.0.0.0",
                "port": 11033
            }
        }
        (changed, input_id) = _check_input_exists(log_inputs, new_input)
        self.assertTrue(input_id is None)


class TestHealthCheck(unittest.TestCase):

    @mock.patch('charms.layer.graylog.api.GraylogApi')
    def test_journal_state(self, mock_api):
        ucommitted_tests = [
            ['test_ok', 0, check_graylog_health.STATE_OK],
            ['test_warning', 1, check_graylog_health.STATE_WARNING],
            ['test_critical', 100, check_graylog_health.STATE_CRITICAL],
        ]
        args = mock.MagicMock(uncommitted_warning=0, uncommitted_critical=100)
        for name, ucommitted, exp_state in ucommitted_tests:
            with self.subTest(name=name):
                mock_api.journal_state.return_value = {
                    'uncommitted_journal_entries': ucommitted}
                state, _ = check_graylog_health.journal_state(mock_api, args)
                self.assertEqual(state, exp_state)


etc_default_conf = u"""
# Path to the java executable.
JAVA=/usr/bin/java

# Default Java options for heap and garbage collection.
{}

# Pass some extra args to graylog-server. (i.e. "-d" to enable debug mode)
GRAYLOG_SERVER_ARGS=""

# Program that will be used to wrap the graylog-server command. Useful to
# support programs like authbind.
GRAYLOG_COMMAND_WRAPPER=""
"""

heap_default = u"""
GRAYLOG_SERVER_JAVA_OPTS="-Xms1g -Xmx1g -XX:NewRatio=1 -server -XX:+ResizeTLAB -XX:+UseConcMarkSweepGC \
-XX:+CMSConcurrentMTEnabled -XX:+CMSClassUnloadingEnabled -XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow"
"""

heap_2g = u"""
GRAYLOG_SERVER_JAVA_OPTS="-Xms2g -Xmx2g -XX:NewRatio=1 -server -XX:+ResizeTLAB -XX:+UseConcMarkSweepGC \
-XX:+CMSConcurrentMTEnabled -XX:+CMSClassUnloadingEnabled -XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow"
"""

heap_not_present = u"""
GRAYLOG_SERVER_JAVA_OPTS="-XX:NewRatio=1 -server -XX:+ResizeTLAB -XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled \
-XX:+CMSClassUnloadingEnabled -XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow -Xms4g -Xmx4g"
"""

heap_unset = u"""
GRAYLOG_SERVER_JAVA_OPTS="-XX:NewRatio=1 -server -XX:+ResizeTLAB -XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled \
-XX:+CMSClassUnloadingEnabled -XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow"
"""

heap_not_present = u"""
GRAYLOG_SERVER_JAVA_OPTS="-XX:NewRatio=1 -server -XX:+ResizeTLAB -XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled \
-XX:+CMSClassUnloadingEnabled -XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow -Xms4g -Xmx4g"
"""

heap_twice = u"""
GRAYLOG_SERVER_JAVA_OPTS="-XX:NewRatio=1 -server -XX:+ResizeTLAB -XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled \
-XX:+CMSClassUnloadingEnabled -XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow -Xms8g -Xmx8g"
"""


class TestSetJVMHeapSize(unittest.TestCase):

    def setUp(self):
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        temp_file.write(etc_default_conf.format(heap_default).encode())
        temp_file.close()
        self.conf_file = temp_file.name

    def tearDown(self):
        os.remove(self.conf_file)

    def conf_equals(self, want):
        with open(self.conf_file, 'r') as conf:
            got = conf.read()
            if got == want:
                return True
            print('{}\n--- != ---\n{}'.format(got, want))
        return False

    def test_same(self):
        set_jvm_heap_size('1g', self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_default)))

    def test_update(self):
        set_jvm_heap_size('2g', self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_2g)))

    def test_unset(self):
        # Remove Xms and Xmx
        set_jvm_heap_size('', self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_unset)))

    def test_add(self):
        set_jvm_heap_size('', self.conf_file)
        set_jvm_heap_size('4g', self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_not_present)))

    def test_add_twice(self):
        set_jvm_heap_size('', self.conf_file)
        set_jvm_heap_size('6g', self.conf_file)
        set_jvm_heap_size('8g', self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_twice)))
