import unittest
from unittest import mock

from charms.layer.graylog import GraylogPipelines, GraylogRules, GraylogStreams, LogExtractPipeline


testdata_simple_pipeline_source = '\npipeline "testpipe"\nstage 1 match either\n  rule "test-rule";\nend\n'

testdata_simple_pipeline = [{'description': None, 'stages':
                             [{'match_all': False, 'rules':
                               ['test-rule'], 'stage': 1}],
                             'source': testdata_simple_pipeline_source,
                             'created_at': '2018-06-18T14:49:27.427Z',
                             'modified_at': '2018-06-18T14:49:27.427Z',
                             'id': '5b27c6778ac7b4647301a562',
                             'errors': None,
                             'title': 'juju-logextract-pipeline'}]

testdata_streams = {'total': 1,
                    'streams': [
                        {'disabled': False, 'remove_matches_from_default_stream': False, 'alert_conditions': [],
                         'created_at': '2018-06-15T21:44:16.285Z', 'id': '000000000000000000000001',
                         'matching_type': 'AND', 'content_pack': None, 'title': 'All messages',
                         'description': 'Stream containing all messages', 'is_default': True,
                         'index_set_id': '5b2433308ac7b463cc19d2ba', 'rules': [], 'creator_user_id': 'local:admin',
                         'alert_receivers': {'emails': [], 'users': []}, 'outputs': []}]}

testdata_rules = [
  {
    "title": "from ceph mon",
    "description": "",
    "source": "rule \"from ceph mon\"\nwhen\n  has_field(\"logdir\") && $message.logdir == \"ceph\" && regex(\"^ceph-mon\", to_string($message.juju_principal_unit)).matches == true\nthen\n  set_field(\"subsystem\", \"ceph-mon\");\nend;\n",  # NOQA: E501
    "created_at": "2018-06-17T14:32:13.632Z",
    "modified_at": "2018-06-17T14:42:25.818Z",
    "errors": None,
    "id": "5b2670ed6140b943aef99ae9"
  }]

testdata_stream_connect = {
    "stream_id": "000000000000000000000001",
    "pipeline_ids": ["5b35061c6140b943ae096769"]
}


class TestGraylogPipelines(unittest.TestCase):

    @mock.patch('charms.layer.graylog.api.GraylogApi')
    def test_get_pipeline_for(self, mock_api):
        mock_api.request.return_value = testdata_simple_pipeline
        gpipes = GraylogPipelines(mock_api)
        test = gpipes.get_for_title('juju-logextract-pipeline')
        self.assertEqual(test['title'], 'juju-logextract-pipeline')

    @mock.patch('charms.layer.graylog.api.GraylogApi')
    def test_set_pipeline(self, mock_api):
        mock_api.request.return_value = testdata_simple_pipeline
        gpipes = GraylogPipelines(mock_api)
        gpipes.set('juju-logextract-pipeline', testdata_simple_pipeline_source)
        self.assertEqual(mock_api.request.call_args_list[1][1]['data']['title'], 'juju-logextract-pipeline')


class TestGraylogRules(unittest.TestCase):

    @mock.patch('charms.layer.graylog.api.GraylogApi')
    def test_get_rules(self, mock_api):
        mock_api.request.return_value = testdata_rules
        grules = GraylogRules(mock_api)
        rules = grules.get()
        self.assertEqual(rules[0]['title'], "from ceph mon")


class TestGraylogStreams(unittest.TestCase):

    @mock.patch('charms.layer.graylog.api.GraylogApi')
    def test_get_stream_id_for(self, mock_api):
        mock_api.request.return_value = testdata_streams
        gstreams = GraylogStreams(mock_api)
        allmsgs = gstreams.get_stream_id_for("All messages")
        self.assertEqual(allmsgs, testdata_streams['streams'][0]['id'])
        nope = gstreams.get_stream_id_for("I don't exist")
        self.assertIsNone(nope)


class TestLogExtractPipeline(unittest.TestCase):

    @mock.patch('charms.layer.graylog.api.GraylogApi')
    def test_get_rules(self, mock_api):
        mock_api.request.return_value = testdata_simple_pipeline

    def test_parse_rule_title(self):
        logext = LogExtractPipeline()
        t = logext.parse_rule_title(testdata_rules[0]['source'])
        self.assertEqual(t, 'from ceph mon')

    @mock.patch('charms.layer.graylog.api.GraylogApi')
    def test_set_rules(self, mock_api):
        logext = LogExtractPipeline(mock_api)
        logext.set_rules([testdata_rules[0]['source']])
        exp = [mock.call('/plugins/org.graylog.plugins.pipelineprocessor/system/pipelines/pipeline')]
        mock_api.request.assert_has_calls(exp)
